"""Module to test the StoreGenericOrderService class."""

import datetime as dt
import math
from copy import deepcopy


from cg.constants import DataDelivery, Priority, Workflow
from cg.constants.constants import CaseActions
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.store_order_services.store_case_order import StoreCaseOrderService
from cg.store.models import Case, Sample
from cg.store.store import Store


def test_cases_to_status(
    mip_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
):
    # GIVEN a scout order with a trio case
    project: OrderType = OrderType.MIP_DNA
    order = OrderIn.parse_obj(mip_order_to_submit, project=project)

    # WHEN parsing for status
    data = store_generic_order_service.order_to_status(order=order)

    # THEN it should pick out the case
    assert len(data["families"]) == 2
    family = data["families"][0]
    assert family["name"] == "family1"
    assert family["data_analysis"] == Workflow.MIP_DNA
    assert family["data_delivery"] == str(DataDelivery.SCOUT)
    assert family["priority"] == Priority.standard.name
    assert family["cohorts"] == ["Other"]
    assert (
        family["synopsis"]
        == "As for the synopsis it will be this overly complex sentence to prove that the synopsis field might in fact be a very long string, which we should be prepared for."
    )
    assert set(family["panels"]) == {"IEM"}
    assert len(family["samples"]) == 3

    first_sample = family["samples"][0]
    assert math.isclose(first_sample["age_at_sampling"], 17.18192, rel_tol=1e-09, abs_tol=1e-09)
    assert first_sample["name"] == "sample1"
    assert first_sample["application"] == "WGSPCFC030"
    assert first_sample["phenotype_groups"] == ["Phenotype-group"]
    assert first_sample["phenotype_terms"] == ["HP:0012747", "HP:0025049"]
    assert first_sample["sex"] == "female"
    assert first_sample["status"] == "affected"
    assert first_sample["subject_id"] == "subject1"
    assert first_sample["mother"] == "sample2"
    assert first_sample["father"] == "sample3"

    # ... second sample has a comment
    assert isinstance(family["samples"][1]["comment"], str)


def test_cases_to_status_synopsis(
    mip_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
):
    # GIVEN a scout order with a trio case where synopsis is None
    modified_order: dict = deepcopy(mip_order_to_submit)
    for sample in modified_order["samples"]:
        sample["synopsis"] = None

    project: OrderType = OrderType.MIP_DNA
    order = OrderIn.parse_obj(mip_order_to_submit, project=project)

    # WHEN parsing for status
    store_generic_order_service.order_to_status(order=order)

    # THEN No exception should have been raised on synopsis


def test_store_mip(
    base_store: Store,
    mip_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a basic store with no samples or nothing in it + scout order
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()

    # WHEN storing the order
    new_families = store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    # THEN it should create and link samples and the case
    assert len(new_families) == 2
    new_case = new_families[0]
    assert new_case.name == "family1"
    assert set(new_case.panels) == {"IEM"}
    assert new_case.priority_human == Priority.standard.name

    assert len(new_case.links) == 3
    new_link = new_case.links[0]
    assert new_case.data_analysis == Workflow.MIP_DNA
    assert new_case.data_delivery == str(DataDelivery.SCOUT)
    assert set(new_case.cohorts) == {"Other"}
    assert (
        new_case.synopsis
        == "As for the synopsis it will be this overly complex sentence to prove that the synopsis field might in fact be a very long string, which we should be prepared for."
    )
    assert new_link.status == "affected"
    assert new_link.mother.name == "sample2"
    assert new_link.father.name == "sample3"
    assert new_link.sample.name == "sample1"
    assert new_link.sample.sex == "female"
    assert new_link.sample.application_version.application.tag == "WGSPCFC030"
    assert new_link.sample.is_tumour
    assert isinstance(new_case.links[1].sample.comment, str)

    assert set(new_link.sample.phenotype_groups) == {"Phenotype-group"}
    assert set(new_link.sample.phenotype_terms) == {"HP:0012747", "HP:0025049"}
    assert new_link.sample.subject_id == "subject1"
    assert math.isclose(new_link.sample.age_at_sampling, 17.18192, rel_tol=1e-09, abs_tol=1e-09)


def test_store_mip_rna(
    base_store: Store,
    mip_rna_status_data,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a basic store with no samples or nothing in it + rna order
    rna_application_tag = "RNAPOAR025"
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()
    assert base_store.get_application_by_tag(tag=rna_application_tag)

    # WHEN storing the order
    new_cases = store_generic_order_service.store_items_in_status(
        customer_id=mip_rna_status_data["customer"],
        order=mip_rna_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_rna_status_data["families"],
    )

    # THEN it should create and link samples and the casing
    assert len(new_cases) == 1
    new_casing = new_cases[0]

    assert len(new_casing.links) == 2
    new_link = new_casing.links[0]
    assert new_casing.data_analysis == Workflow.MIP_RNA
    assert new_casing.data_delivery == str(DataDelivery.SCOUT)
    assert new_link.sample.name == "sample1-rna-t1"
    assert new_link.sample.application_version.application.tag == rna_application_tag


def test_store_cancer_samples(
    base_store: Store,
    balsamic_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):

    # GIVEN a basic store with no samples and a cancer order
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()

    # WHEN storing the order
    new_families = store_generic_order_service.store_items_in_status(
        customer_id=balsamic_status_data["customer"],
        order=balsamic_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=balsamic_status_data["families"],
    )

    # THEN it should create and link samples and the case
    assert len(new_families) == 1
    new_case = new_families[0]
    assert new_case.name == "family1"
    assert new_case.data_analysis in [
        Workflow.BALSAMIC,
        Workflow.BALSAMIC_QC,
        Workflow.BALSAMIC_UMI,
    ]
    assert new_case.data_delivery == str(DataDelivery.FASTQ_ANALYSIS_SCOUT)
    assert not set(new_case.panels)
    assert new_case.priority_human == Priority.standard.name

    assert len(new_case.links) == 1
    new_link = new_case.links[0]
    assert new_link.sample.name == "s1"
    assert new_link.sample.sex == "male"
    assert new_link.sample.application_version.application.tag == "WGSPCFC030"
    assert new_link.sample.comment == "other Elution buffer"
    assert new_link.sample.is_tumour


def test_store_existing_single_sample_from_trio(
    base_store: Store,
    mip_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a stored trio case

    new_families = store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    new_case = new_families[0]
    assert new_case.name == "family1"
    assert set(new_case.panels) == {"IEM"}
    assert new_case.priority_human == Priority.standard.name

    assert len(new_case.links) == 3
    new_link = new_case.links[0]
    assert new_link.mother
    assert new_link.father
    name = new_link.sample.name
    internal_id = new_link.sample.internal_id
    assert base_store.get_sample_by_internal_id(internal_id)

    # WHEN storing a new case with one sample from the trio
    for family_idx, family in enumerate(mip_status_data["families"]):
        for sample_idx, sample in enumerate(family["samples"]):
            if sample["name"] == name:
                sample["internal_id"] = internal_id
                family["name"] = "single-from-trio"
            else:
                family["samples"][sample_idx] = {}

        family["samples"] = list(filter(None, family["samples"]))

        if not family["samples"]:
            mip_status_data["families"][family_idx] = {}

    mip_status_data["families"] = list(filter(None, mip_status_data["families"]))

    new_families = store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    # THEN there should be no complaints about missing parents
    assert len(new_families) == 1
    assert len(new_families[0].links) == 1
    assert not new_families[0].links[0].mother
    assert not new_families[0].links[0].father


def test_store_existing_case(
    base_store: Store,
    mip_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a basic store with no samples or nothing in it + scout order
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()

    # WHEN storing the order
    store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    base_store.session.close()
    new_cases: list[Case] = base_store.get_cases()

    # Save internal id
    stored_cases_internal_ids = dict([(case.name, case.internal_id) for case in new_cases])
    for case in mip_status_data["families"]:
        case["internal_id"] = stored_cases_internal_ids[case["name"]]

    store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    base_store.session.close()
    rerun_cases: list[Case] = base_store.get_cases()

    # THEN the sample ticket should be appended to previos ticket and action set to analyze
    assert rerun_cases[0].tickets == f"{ticket_id},{ticket_id}"
    assert rerun_cases[0].action == CaseActions.ANALYZE



import logging

from cg.services.run_devices.abstract_classes import PostProcessingStoreService
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingDataTransferError,
    PostProcessingStoreDataError,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioDTOs,
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.store.models import PacBioSequencingRun, PacBioSMRTCell
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class PacBioStoreService(PostProcessingStoreService):
    def __init__(self, store: Store, data_transfer_service: PacBioDataTransferService):
        self.store = store
        self.data_transfer_service = data_transfer_service

    def _create_run_device(self, run_device_dto: PacBioSMRTCellDTO) -> PacBioSMRTCell:
        return self.store.create_pac_bio_smrt_cell(run_device_dto)

    def _create_instrument_run(
        self, instrument_run_dto: PacBioSequencingRunDTO, smrt_cell: PacBioSMRTCell
    ) -> PacBioSequencingRun:
        return self.store.create_pac_bio_sequencing_run(
            sequencing_run_dto=instrument_run_dto, smrt_cell=smrt_cell
        )

    def _create_sample_run_metrics(
        self,
        sample_run_metrics_dtos: list[PacBioSampleSequencingMetricsDTO],
        sequencing_run: PacBioSequencingRun,
    ) -> None:
        for sample_run_metric in sample_run_metrics_dtos:
            self.store.create_pac_bio_sample_sequencing_run(
                sample_run_metrics_dto=sample_run_metric, sequencing_run=sequencing_run
            )

    @handle_post_processing_errors(
        to_except=(PostProcessingDataTransferError, ValueError),
        to_raise=PostProcessingStoreDataError,
    )
    def store_post_processing_data(self, run_data: PacBioRunData, dry_run: bool = False) -> None:
        dtos: PacBioDTOs = self.data_transfer_service.get_post_processing_dtos(run_data)
        smrt_cell: PacBioSMRTCell = self._create_run_device(dtos.run_device)
        sequencing_run: PacBioSequencingRun = self._create_instrument_run(
            instrument_run_dto=dtos.sequencing_run, smrt_cell=smrt_cell
        )
        self._create_sample_run_metrics(
            sample_run_metrics_dtos=dtos.sample_sequencing_metrics, sequencing_run=sequencing_run
        )
        if dry_run:
            self.store.rollback()
            LOG.info(
                f"Dry run, no entries will be added to database for SMRT cell {run_data.full_path}."
            )
            return
        self.store.commit_to_store()



from cg.exc import OrderError
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import Of1508Sample, OrderInSample
from cg.services.orders.submitters.order_submitter import ValidateOrderService
from cg.store.models import Sample, Customer
from cg.store.store import Store


class ValidateCaseOrderService(ValidateOrderService):

    def __init__(self, status_db: Store):
        self.status_db = status_db

    def validate_order(self, order: OrderIn) -> None:
        self._validate_subject_sex(samples=order.samples, customer_id=order.customer)
        self._validate_samples_available_to_customer(
            samples=order.samples, customer_id=order.customer
        )
        self._validate_case_names_are_unique(samples=order.samples, customer_id=order.customer)
        if order.order_type == OrderType.RNAFUSION:
            self._validate_only_one_sample_per_case(samples=order.samples)

    def _validate_subject_sex(self, samples: [Of1508Sample], customer_id: str):
        """Validate that sex is consistent with existing samples, skips samples of unknown sex

        Args:
            samples     (list[dict]):   Samples to validate
            customer_id (str):          Customer that the samples belong to
        Returns:
            Nothing
        """
        sample: Of1508Sample
        for sample in samples:
            subject_id: str = sample.subject_id
            if not subject_id:
                continue
            new_gender: str = sample.sex
            if new_gender == "unknown":
                continue

            existing_samples: list[Sample] = self.status_db.get_samples_by_customer_and_subject_id(
                customer_internal_id=customer_id, subject_id=subject_id
            )
            existing_sample: Sample
            for existing_sample in existing_samples:
                previous_gender = existing_sample.sex
                if previous_gender == "unknown":
                    continue

                if previous_gender != new_gender:
                    raise OrderError(
                        f"Sample gender inconsistency for subject_id: {subject_id}: previous gender {previous_gender}, new gender {new_gender}"
                    )

    def _validate_samples_available_to_customer(
        self, samples: list[OrderInSample], customer_id: str
    ) -> None:
        """Validate that the customer have access to all samples"""
        sample: Of1508Sample
        for sample in samples:
            if not sample.internal_id:
                continue

            existing_sample: Sample = self.status_db.get_sample_by_internal_id(
                internal_id=sample.internal_id
            )

            data_customer: Customer = self.status_db.get_customer_by_internal_id(
                customer_internal_id=customer_id
            )

            if existing_sample.customer not in data_customer.collaborators:
                raise OrderError(f"Sample not available: {sample.name}")

    def _validate_case_names_are_unique(
        self, samples: list[OrderInSample], customer_id: str
    ) -> None:
        """Validate that the names of all cases are unused for all samples"""

        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )

        sample: Of1508Sample
        for sample in samples:
            if self._is_rerun_of_existing_case(sample=sample):
                continue
            if self.status_db.get_case_by_name_and_customer(
                customer=customer, case_name=sample.family_name
            ):
                raise OrderError(f"Case name {sample.family_name} already in use")

    @staticmethod
    def _is_rerun_of_existing_case(sample: Of1508Sample) -> bool:
        return sample.case_internal_id is not None

    @staticmethod
    def _validate_only_one_sample_per_case(samples: list[Of1508Sample]) -> None:
        """Validates that each case contains only one sample."""
        if len({sample.family_name for sample in samples}) != len(samples):
            raise OrderError("Each case in an RNAFUSION order must have exactly one sample.")



"""Fixtures for the collect qc metrics api."""

from datetime import datetime
from pathlib import Path

import pytest
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.clients.arnold.api import ArnoldAPIClient
from cg.clients.arnold.dto.create_case_request import CreateCaseRequest
from cg.clients.janus.api import JanusAPIClient
from cg.clients.janus.dto.create_qc_metrics_request import (
    CreateQCMetricsRequest,
    FilePathAndTag,
    WorkflowInfo,
)
from cg.constants.housekeeper_tags import JanusTags
from cg.meta.qc_metrics.collect_qc_metrics import CollectQCMetricsAPI
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def janus_hk_tags() -> list[str]:
    return ["qc-metrics", "janus", "multiqc"]


@pytest.fixture
def file_with_right_tags() -> str:
    return Path("tests", "fixtures", "meta", "qc_metrics", "file_with_right_tags.json").__str__()


@pytest.fixture
def file_without_right_tags() -> str:
    return Path("tests", "fixtures", "meta", "qc_metrics", "file_without_right_tags.json").__str__()


@pytest.fixture(scope="function")
def janus_hk_store(
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    timestamp: datetime,
    case_id_with_single_sample: str,
    file_with_right_tags: str,
    file_without_right_tags: str,
) -> HousekeeperAPI:
    bundle_with_janus_files = {
        "name": case_id_with_single_sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": file_with_right_tags,
                "archive": False,
                "tags": ["qc-metrics", "janus", "multiqc", "picard-hs"],
            },
            {
                "path": file_without_right_tags,
                "archive": False,
                "tags": ["not", "the", "tags", "you", "are", "looking", "for"],
            },
        ],
    }
    helpers.ensure_hk_bundle(real_housekeeper_api, bundle_with_janus_files, include=True)
    return real_housekeeper_api


@pytest.fixture(scope="function")
def janus_store(
    store_with_multiple_cases_and_samples: Store, helpers: StoreHelpers, case_id_with_single_sample
) -> Store:
    case: Case = store_with_multiple_cases_and_samples.get_case_by_internal_id(
        case_id_with_single_sample
    )
    helpers.add_analysis(store_with_multiple_cases_and_samples, case=case)
    return store_with_multiple_cases_and_samples


@pytest.fixture
def mock_janus_api() -> JanusAPIClient:
    return JanusAPIClient(config={"janus": {"host": ""}})


@pytest.fixture
def mock_arnold_api() -> ArnoldAPIClient:
    return ArnoldAPIClient(config={"arnold": {"api_url": ""}})


@pytest.fixture(scope="function")
def collect_qc_metrics_api(
    janus_store: Store,
    janus_hk_store: HousekeeperAPI,
    mock_janus_api: JanusAPIClient,
    mock_arnold_api: ArnoldAPIClient,
):
    return CollectQCMetricsAPI(
        hk_api=janus_hk_store,
        status_db=janus_store,
        janus_api=mock_janus_api,
        arnold_api=mock_arnold_api,
    )


@pytest.fixture
def expected_request(
    case_id_with_single_sample: str,
    sample_id_in_single_case: str,
    janus_hk_store: HousekeeperAPI,
    janus_store: Store,
) -> CreateQCMetricsRequest:
    hk_file: File = janus_hk_store.get_files(
        case_id_with_single_sample, tags=JanusTags.tags_to_retrieve
    ).first()
    sample: Sample = janus_store.get_samples_by_case_id(case_id_with_single_sample)[0]
    case: Case = janus_store.get_case_by_internal_id(case_id_with_single_sample)
    file_path_and_tag = FilePathAndTag(file_path=hk_file.full_path, tag="picard-hs")
    return CreateQCMetricsRequest(
        case_id=case_id_with_single_sample,
        sample_ids=[sample.internal_id],
        workflow_info=WorkflowInfo(
            workflow=case.data_analysis, version=case.analyses[0].workflow_version
        ),
        files=[file_path_and_tag],
    )


@pytest.fixture
def mock_case_id() -> str:
    return "mock_id"


@pytest.fixture
def mock_case_qc_metrics(mock_case_id: str) -> dict:
    return {"case_id": mock_case_id, "case_info": {}}


@pytest.fixture
def expected_create_case_request(mock_case_qc_metrics: dict) -> CreateCaseRequest:
    return CreateCaseRequest(**mock_case_qc_metrics)



from functools import wraps

from cg.exc import CgError
from cg.services.deliver_files.delivery_file_tag_fetcher_service.exc import (
    FetchDeliveryFileTagsError,
)


def handle_tag_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValueError as error:
            raise FetchDeliveryFileTagsError from error
        except Exception as error:
            raise CgError(f"{error}") from error

    return wrapper



from cg.exc import CgError
from cg.models.cg_config import CGConfig
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService
from cg.utils.mapping import get_item_by_pattern_in_source

PATTERN_TO_DEVICE_MAP: dict[str, str] = {
    r"^r\d+_\d+_\d+/(1|2)_[^/]+$": "pacbio",
}


def get_post_processing_service_from_run_name(
    context: CGConfig, run_name: str
) -> PacBioPostProcessingService:
    """Get the correct post-processing service based on the run name."""
    try:
        device: str = get_item_by_pattern_in_source(
            source=run_name, pattern_map=PATTERN_TO_DEVICE_MAP
        )
    except CgError as error:
        raise NameError(
            f"Run name {run_name} does not match with any known sequencing run name pattern"
        ) from error
    return getattr(context.post_processing_services, device)



"""CLI for delivering files with CG"""

import logging
from pathlib import Path

import click

from cg.apps.tb import TrailblazerAPI
from cg.cli.deliver.utils import deliver_raw_data_for_analyses
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants import Workflow
from cg.constants.cli_options import DRY_RUN
from cg.constants.delivery import FileDeliveryOption
from cg.services.deliver_files.delivery_rsync_service.delivery_rsync_service import (
    DeliveryRsyncService,
)
from cg.models.cg_config import CGConfig
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.deliver_files_service_factory import (
    DeliveryServiceFactory,
)
from cg.store.models import Case, Analysis

LOG = logging.getLogger(__name__)

DELIVERY_TYPE = click.option(
    "-d",
    "--delivery-type",
    multiple=False,
    type=click.Choice(choices=[option for option in FileDeliveryOption]),
    required=False,
)
TICKET_ID_ARG = click.option("-t", "--ticket", type=str, required=True)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
def deliver():
    """Deliver files with CG."""
    LOG.info("Running CG deliver")


@deliver.command(name="rsync")
@click.pass_obj
@TICKET_ID_ARG
@DRY_RUN
def rsync(context: CGConfig, ticket: str, dry_run: bool):
    """The folder generated using the "cg deliver analysis" command will be
    rsynced with this function to the customers inbox on the delivery server
    """
    tb_api: TrailblazerAPI = context.trailblazer_api
    rsync_api: DeliveryRsyncService = context.delivery_rsync_service
    slurm_id = rsync_api.run_rsync_for_ticket(ticket=ticket, dry_run=dry_run)
    LOG.info(f"Rsync to the delivery server running as job {slurm_id}")
    rsync_api.add_to_trailblazer_api(
        tb_api=tb_api, slurm_job_id=slurm_id, ticket=ticket, dry_run=dry_run
    )


@deliver.command(
    name="case",
    help="Deliver all case files based on delivery type to the customer inbox on the HPC "
    "and start an Rsync job to clinical-delivery. "
    "NOTE: the dry-run flag will copy files to the customer inbox on Hasta, "
    "but will not perform the Rsync job.",
)
@click.pass_obj
@click.option(
    "-c",
    "--case-id",
    required=True,
    help="Deliver the files for a specific case.",
)
@DELIVERY_TYPE
@DRY_RUN
def deliver_case(
    context: CGConfig,
    case_id: str,
    delivery_type: FileDeliveryOption,
    dry_run: bool,
):
    """
    Deliver all case files based on delivery type to the customer inbox on the HPC
    """
    inbox: str = context.delivery_path
    service_builder: DeliveryServiceFactory = context.delivery_service_factory
    case: Case = context.status_db.get_case_by_internal_id(internal_id=case_id)
    if not case:
        LOG.error(f"Could not find case with id {case_id}")
        return
    delivery_service: DeliverFilesService = service_builder.build_delivery_service(
        delivery_type=delivery_type if delivery_type else case.data_delivery,
        workflow=case.data_analysis,
    )
    delivery_service.deliver_files_for_case(
        case=case, delivery_base_path=Path(inbox), dry_run=dry_run
    )


@deliver.command(
    name="ticket",
    help="Deliver all case files for cases in a ticket based on delivery type to the customer inbox on the HPC "
    "and start an Rsync job to clinical-delivery. "
    "NOTE: the dry-run flag will copy files to the customer inbox on Hasta, "
    "but will not perform the Rsync job.",
)
@click.pass_obj
@TICKET_ID_ARG
@DELIVERY_TYPE
@DRY_RUN
def deliver_ticket(
    context: CGConfig,
    ticket: str,
    delivery_type: FileDeliveryOption,
    dry_run: bool,
):
    """
    Deliver all case files based on delivery type to the customer inbox on the HPC for cases connected to a ticket.
    """
    inbox: str = context.delivery_path
    service_builder: DeliveryServiceFactory = context.delivery_service_factory
    cases: list[Case] = context.status_db.get_cases_by_ticket_id(ticket_id=ticket)
    if not cases:
        LOG.error(f"Could not find case connected to ticket {ticket}")
        return
    delivery_service: DeliverFilesService = service_builder.build_delivery_service(
        delivery_type=delivery_type if delivery_type else cases[0].data_delivery,
        workflow=cases[0].data_analysis,
    )
    delivery_service.deliver_files_for_ticket(
        ticket_id=ticket, delivery_base_path=Path(inbox), dry_run=dry_run
    )


@deliver.command(name="auto-raw-data")
@click.pass_obj
@DRY_RUN
def deliver_auto_raw_data(context: CGConfig, dry_run: bool):
    """
    Deliver all case files for the raw data workflow to the customer inbox on the HPC and start a Rsync job.
    1. get all cases with analysis type fastq that need to be delivered
    2. check if their upload has started
    3. if not, start the upload
    4. update the uploaded at field
    """
    """Starts upload of all not previously uploaded cases with analysis type fastq to
       clinical-delivery."""
    service_builder: DeliveryServiceFactory = context.delivery_service_factory
    analyses: list[Analysis] = context.analysis_service.get_analyses_to_upload_for_workflow(
        workflow=Workflow.FASTQ
    )
    deliver_raw_data_for_analyses(
        analyses=analyses,
        status_db=context.status_db,
        delivery_path=Path(context.delivery_path),
        service_builder=service_builder,
        dry_run=dry_run,
    )



"""Module for the PacBioHousekeeperService tests."""

from pathlib import Path
from unittest import mock

import pytest
from housekeeper.store.models import File

from cg.services.run_devices.exc import (
    PostProcessingParsingError,
    PostProcessingRunFileManagerError,
    PostProcessingStoreFileError,
)
from cg.services.run_devices.pacbio.housekeeper_service.models import PacBioFileData
from cg.services.run_devices.pacbio.housekeeper_service.pacbio_houskeeper_service import (
    PacBioHousekeeperService,
)
from cg.services.run_devices.pacbio.metrics_parser.metrics_parser import PacBioMetricsParser
from cg.services.run_devices.pacbio.metrics_parser.models import PacBioMetrics
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager


def test_store_files_in_housekeeper(
    pac_bio_housekeeper_service: PacBioHousekeeperService,
    expected_pac_bio_run_data: PacBioRunData,
    pac_bio_sample_internal_id: str,
    smrt_cell_internal_id: str,
):
    # GIVEN a PacBioRunData object and a PacBioHousekeeperService

    # WHEN storing files in Housekeeper
    pac_bio_housekeeper_service.store_files_in_housekeeper(expected_pac_bio_run_data)

    # THEN a SMRT cell bundle is created
    assert pac_bio_housekeeper_service.hk_api.get_latest_bundle_version(smrt_cell_internal_id)

    # THEN a sample bundle is created
    assert pac_bio_housekeeper_service.hk_api.get_latest_bundle_version(pac_bio_sample_internal_id)

    # THEN all expected files are listed under the smrt cell bundle
    smrt_bundle_files: list[File] = (
        pac_bio_housekeeper_service.hk_api.get_files_from_latest_version(
            bundle_name=smrt_cell_internal_id
        )
    )
    assert smrt_bundle_files

    # THEN all expected files are listed under the sample bundle
    sample_bundle_files: list[File] = pac_bio_housekeeper_service.hk_api.get_latest_bundle_version(
        bundle_name=pac_bio_sample_internal_id
    ).files
    assert sample_bundle_files


@pytest.mark.parametrize(
    "file_fixture, file_data_fixture",
    [
        ("pac_bio_ccs_report_file", "ccs_report_pac_bio_file_data"),
        ("pac_bio_hifi_read_file", "pac_bio_hifi_reads_file_data"),
    ],
    ids=["ccs_report", "hifi_reads"],
)
def test_create_bundle_info(
    pac_bio_housekeeper_service: PacBioHousekeeperService,
    pac_bio_metrics: PacBioMetrics,
    file_fixture: str,
    file_data_fixture: str,
    request: pytest.FixtureRequest,
):
    """Test the correct file data object is created for a file."""
    # GIVEN a PacBioMetrics and a Housekeeper PacBioHousekeeperService objects

    # GIVEN a file path
    file: Path = request.getfixturevalue(file_fixture)

    # WHEN creating the file data object for the file
    file_data: PacBioFileData = pac_bio_housekeeper_service._create_bundle_info(
        file_path=file, parsed_metrics=pac_bio_metrics
    )

    # THEN the method should return the correct PacBioFileData object
    expected_file_data: PacBioFileData = request.getfixturevalue(file_data_fixture)
    assert file_data == expected_file_data


def test_store_housekeeper_file_not_found(
    pac_bio_housekeeper_service: PacBioHousekeeperService,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a PacBioRunData object and a PacBioHousekeeperService

    # WHEN storing files in Housekeeper

    # THEN an error is raised
    with mock.patch.object(
        PacBioRunFileManager, "get_files_to_store", side_effect=PostProcessingRunFileManagerError
    ):
        with pytest.raises(PostProcessingStoreFileError):
            pac_bio_housekeeper_service.store_files_in_housekeeper(expected_pac_bio_run_data)


def test_store_housekeeper_parsing_error(
    pac_bio_housekeeper_service: PacBioHousekeeperService,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a PacBioRunData object and a PacBioHousekeeperService

    # WHEN storing files in Housekeeper

    # THEN an error is raised
    with mock.patch.object(
        PacBioMetricsParser, "parse_metrics", side_effect=PostProcessingParsingError
    ):
        with pytest.raises(PostProcessingStoreFileError):
            pac_bio_housekeeper_service.store_files_in_housekeeper(expected_pac_bio_run_data)



import datetime as dt
import pytest
from cg.constants import DataDelivery, Workflow
from cg.constants.constants import PrepCategory
from cg.exc import OrderError
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.store_order_services.store_fastq_order_service import StoreFastqOrderService
from cg.store.models import Application, Case, Sample
from cg.store.store import Store


def test_samples_to_status(
    fastq_order_to_submit: dict, store_fastq_order_service: StoreFastqOrderService
):
    # GIVEN fastq order with two samples
    order = OrderIn.parse_obj(fastq_order_to_submit, OrderType.FASTQ)

    # WHEN parsing for status
    data = store_fastq_order_service.order_to_status(order=order)

    # THEN it should pick out samples and relevant information
    assert len(data["samples"]) == 2
    first_sample = data["samples"][0]
    assert first_sample["name"] == "prov1"
    assert first_sample["application"] == "WGSPCFC060"
    assert first_sample["priority"] == "priority"
    assert first_sample["tumour"] is False
    assert first_sample["volume"] == "1"

    # ... and the other sample is a tumour
    assert data["samples"][1]["tumour"] is True


def test_store_samples(
    base_store: Store,
    fastq_status_data: dict,
    ticket_id: str,
    store_fastq_order_service: StoreFastqOrderService,
):
    # GIVEN a basic store with no samples and a fastq order
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0

    # WHEN storing the order
    new_samples = store_fastq_order_service.store_items_in_status(
        customer_id=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=fastq_status_data["samples"],
    )

    # THEN it should store the samples and create a case for each sample
    assert len(new_samples) == 2
    assert len(base_store._get_query(table=Sample).all()) == 2
    assert base_store._get_query(table=Case).count() == 2
    first_sample = new_samples[0]
    assert len(first_sample.links) == 2
    family_link = first_sample.links[0]
    assert family_link.case in base_store.get_cases()
    assert family_link.case.data_analysis
    assert family_link.case.data_delivery in [DataDelivery.FASTQ, DataDelivery.NO_DELIVERY]


def test_store_samples_sex_stored(
    base_store: Store,
    fastq_status_data: dict,
    ticket_id: str,
    store_fastq_order_service: StoreFastqOrderService,
):
    # GIVEN a basic store with no samples and a fastq order
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0

    # WHEN storing the order
    new_samples = store_fastq_order_service.store_items_in_status(
        customer_id=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=fastq_status_data["samples"],
    )

    # THEN the sample sex should be stored
    assert new_samples[0].sex == "male"


def test_store_fastq_samples_non_tumour_wgs_to_mip(
    base_store: Store, fastq_status_data: dict, store_fastq_order_service: StoreFastqOrderService
):
    # GIVEN a basic store with no samples and a non-tumour fastq order as wgs
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0
    base_store.get_application_by_tag(
        fastq_status_data["samples"][0]["application"]
    ).prep_category = PrepCategory.WHOLE_GENOME_SEQUENCING
    fastq_status_data["samples"][0]["tumour"] = False

    # WHEN storing the order
    new_samples = store_fastq_order_service.store_items_in_status(
        customer_id=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=1234348,
        items=fastq_status_data["samples"],
    )

    # THEN the analysis for the case should be MAF
    assert new_samples[0].links[0].case.data_analysis == Workflow.MIP_DNA


def test_store_fastq_samples_tumour_wgs_to_fastq(
    base_store: Store,
    fastq_status_data: dict,
    ticket_id: str,
    store_fastq_order_service: StoreFastqOrderService,
):
    # GIVEN a basic store with no samples and a tumour fastq order as wgs
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0
    base_store.get_application_by_tag(
        fastq_status_data["samples"][0]["application"]
    ).prep_category = PrepCategory.WHOLE_GENOME_SEQUENCING
    fastq_status_data["samples"][0]["tumour"] = True

    # WHEN storing the order
    new_samples = store_fastq_order_service.store_items_in_status(
        customer_id=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=fastq_status_data["samples"],
    )

    # THEN the analysis for the case should be FASTQ
    assert new_samples[0].links[0].case.data_analysis == Workflow.FASTQ


def test_store_fastq_samples_non_wgs_as_fastq(
    base_store: Store,
    fastq_status_data: dict,
    ticket_id: str,
    store_fastq_order_service: StoreFastqOrderService,
):
    # GIVEN a basic store with no samples and a fastq order as non wgs
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0
    non_wgs_prep_category = PrepCategory.WHOLE_EXOME_SEQUENCING

    non_wgs_applications = base_store._get_query(table=Application).filter(
        Application.prep_category == non_wgs_prep_category
    )

    assert non_wgs_applications

    for sample in fastq_status_data["samples"]:
        sample["application"] = non_wgs_applications[0].tag

    # WHEN storing the order
    new_samples = store_fastq_order_service.store_items_in_status(
        customer_id=fastq_status_data["customer"],
        order=fastq_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=fastq_status_data["samples"],
    )

    # THEN the analysis for the case should be fastq (none)
    assert new_samples[0].links[0].case.data_analysis == Workflow.FASTQ


def test_store_samples_bad_apptag(
    base_store: Store,
    fastq_status_data: dict,
    ticket_id: str,
    store_fastq_order_service: StoreFastqOrderService,
):
    # GIVEN a basic store with no samples and a fastq order
    assert not base_store._get_query(table=Sample).first()
    assert base_store._get_query(table=Case).count() == 0

    for sample in fastq_status_data["samples"]:
        sample["application"] = "nonexistingtag"

    # THEN it should raise OrderError
    with pytest.raises(OrderError):
        # WHEN storing the order
        store_fastq_order_service.store_items_in_status(
            customer_id=fastq_status_data["customer"],
            order=fastq_status_data["order"],
            ordered=dt.datetime.now(),
            ticket_id=ticket_id,
            items=fastq_status_data["samples"],
        )



import pytest

from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.services.orders.store_order_services.store_fastq_order_service import StoreFastqOrderService
from cg.services.orders.store_order_services.store_case_order import StoreCaseOrderService
from cg.services.orders.store_order_services.store_metagenome_order import (
    StoreMetagenomeOrderService,
)
from cg.services.orders.store_order_services.store_microbial_fastq_order_service import (
    StoreMicrobialFastqOrderService,
)
from cg.services.orders.store_order_services.store_microbial_order import StoreMicrobialOrderService
from cg.services.orders.store_order_services.store_pool_order import StorePoolOrderService
from cg.store.store import Store
from tests.mocks.limsmock import MockLimsAPI


@pytest.fixture
def store_generic_order_service(base_store: Store, lims_api: MockLimsAPI) -> StoreCaseOrderService:
    return StoreCaseOrderService(status_db=base_store, lims_service=OrderLimsService(lims_api))


@pytest.fixture
def store_pool_order_service(base_store: Store, lims_api: MockLimsAPI) -> StorePoolOrderService:
    return StorePoolOrderService(status_db=base_store, lims_service=OrderLimsService(lims_api))


@pytest.fixture
def store_fastq_order_service(base_store: Store, lims_api: MockLimsAPI) -> StoreFastqOrderService:
    return StoreFastqOrderService(status_db=base_store, lims_service=OrderLimsService(lims_api))


@pytest.fixture
def store_metagenome_order_service(
    base_store: Store, lims_api: MockLimsAPI
) -> StoreMetagenomeOrderService:
    return StoreMetagenomeOrderService(
        status_db=base_store, lims_service=OrderLimsService(lims_api)
    )


@pytest.fixture
def store_microbial_order_service(
    base_store: Store, lims_api: MockLimsAPI
) -> StoreMicrobialOrderService:
    return StoreMicrobialOrderService(status_db=base_store, lims_service=OrderLimsService(lims_api))


@pytest.fixture
def store_microbial_fastq_order_service(
    base_store: Store, lims_api: MockLimsAPI
) -> StoreMicrobialFastqOrderService:
    return StoreMicrobialFastqOrderService(
        status_db=base_store, lims_service=OrderLimsService(lims_api)
    )



"""Tests CLI common methods to create the case config for NF analyses."""

import logging
from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.apps.lims import LimsAPI
from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.constants.constants import FileFormat, MetaApis
from cg.constants.nextflow import NEXTFLOW_WORKFLOWS
from cg.io.controller import ReadFile
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.utils import Process

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_config_case_without_options(
    cli_runner: CliRunner, workflow: Workflow, request: FixtureRequest
):
    """Test config_case for workflow without options."""
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # WHEN invoking the command without additional parameters
    result = cli_runner.invoke(workflow_cli, [workflow, "config-case"], obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command log should inform about missing arguments
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_config_with_missing_case(
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    workflow: Workflow,
    request: FixtureRequest,
):
    """Test config_case for workflow with a missing case."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case not in the StatusDB database
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN running
    result = cli_runner.invoke(
        workflow_cli, [workflow, "config-case", case_id_does_not_exist], obj=context
    )

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN the error log should indicate that the case is invalid
    assert "could not be found in Status DB!" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_config_case_without_samples(
    cli_runner: CliRunner,
    workflow: Workflow,
    no_sample_case_id: str,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
):
    """Test command with a case lacking linked samples."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case without linked samples

    # WHEN invoking the command
    result = cli_runner.invoke(
        workflow_cli, [workflow, "config-case", no_sample_case_id], obj=context
    )

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN warning should be printed that no sample is found
    assert no_sample_case_id in caplog.text
    assert "has no samples" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_config_case_default_parameters(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
    scout_export_manged_variants_output: str,
    mocker,
):
    """Test that command generates config files."""
    caplog.set_level(logging.DEBUG)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    sample_sheet_path: Path = request.getfixturevalue(f"{workflow}_sample_sheet_path")
    params_file_path: Path = request.getfixturevalue(f"{workflow}_params_file_path")
    nexflow_config_file_path: Path = request.getfixturevalue(f"{workflow}_nexflow_config_file_path")
    sample_sheet_content_expected: str = request.getfixturevalue(f"{workflow}_sample_sheet_content")

    # Mocking external Scout call
    mocker.patch.object(Process, "run_command", return_value=None)

    # GIVEN that the sample source in LIMS is set
    mocker.patch.object(LimsAPI, "get_source", return_value="blood")

    # GIVEN a mocked scout export of the managed variants
    mocker.patch.object(
        RarediseaseAnalysisAPI,
        "get_managed_variants",
        return_value=scout_export_manged_variants_output,
    )

    # GIVEN a valid case

    # WHEN running config case
    result = cli_runner.invoke(workflow_cli, [workflow, "config-case", case_id], obj=context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN logs should be as expected
    expected_logs: list[str] = [
        "Getting sample sheet information",
        "Writing sample sheet",
        "Getting parameters information",
        "Writing parameters file",
        "Writing nextflow config file",
    ]
    for expected_log in expected_logs:
        assert expected_log in caplog.text

    # THEN config files should be generated
    assert sample_sheet_path.is_file()
    assert params_file_path.is_file()
    assert nexflow_config_file_path.is_file()

    # THEN the sample sheet content should match the expected values
    sample_sheet_content_created: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.TXT, file_path=sample_sheet_path, read_to_string=True
    )
    assert sample_sheet_content_expected in sample_sheet_content_created

    # THEN the params file should contain all parameters
    parameters_default = vars(request.getfixturevalue(f"{workflow}_parameters_default"))
    params_content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.TXT, file_path=params_file_path, read_to_string=True
    )
    for parameter in parameters_default:
        assert parameter in params_content

    # WHEN the workflow requires a gene panel
    # THEN information about the panel being generated should be logged and the file should be written
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    if analysis_api.is_gene_panel_required:
        assert "Creating gene panel file" in caplog.text
        gene_panel_path: Path = request.getfixturevalue(f"{workflow}_gene_panel_path")
        assert gene_panel_path.is_file()


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_config_case_dry_run(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
    scout_export_manged_variants_output: str,
    mocker,
):
    """Test dry-run."""
    caplog.set_level(logging.DEBUG)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    sample_sheet_path: Path = request.getfixturevalue(f"{workflow}_sample_sheet_path")
    params_file_path: Path = request.getfixturevalue(f"{workflow}_params_file_path")
    nexflow_config_file_path: Path = request.getfixturevalue(f"{workflow}_nexflow_config_file_path")

    # GIVEN a valid case

    # GIVEN that the sample source in LIMS is set
    mocker.patch.object(LimsAPI, "get_source", return_value="blood")

    # GIVEN a mocked scout export of the managed variants
    mocker.patch.object(
        RarediseaseAnalysisAPI,
        "get_managed_variants",
        return_value=scout_export_manged_variants_output,
    )

    # WHEN invoking the command with dry-run specified
    result = cli_runner.invoke(
        workflow_cli, [workflow, "config-case", case_id, "--dry-run"], obj=context
    )

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN sample sheet and parameters information should be collected
    assert "Getting sample sheet information" in caplog.text
    assert "Getting parameters information" in caplog.text

    # THEN sample sheet and parameters information files should not be written
    assert "Dry run: Config files will not be written" in caplog.text
    assert not sample_sheet_path.is_file()
    assert not params_file_path.is_file()
    assert not nexflow_config_file_path.is_file()

    # WHEN the workflow requires a gene panel
    # THEN information about the panel being generated should be logged but no file should be written
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    if analysis_api.is_gene_panel_required:
        assert "Creating gene panel file" in caplog.text
        assert "bin/scout --config scout-stage.yaml export panel" in caplog.text
        gene_panel_path: Path = request.getfixturevalue(f"{workflow}_gene_panel_path")
        assert not gene_panel_path.is_file()



"""Filters for the Illumina Flow Cells."""

from enum import Enum

from sqlalchemy.orm import Query


def filter_illumina_flow_cell_by_internal_id(
    flow_cells: Query, internal_id: str, **kwargs
) -> Query:
    """Get an Illumina Flow Cell by its internal id."""
    return flow_cells.filter_by(internal_id=internal_id)


def apply_illumina_flow_cell_filters(
    flow_cells: Query, filter_functions: list[callable], internal_id: str
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        flow_cells: Query = function(
            flow_cells=flow_cells,
            internal_id=internal_id,
        )
    return flow_cells


class IlluminaFlowCellFilter(Enum):
    """Define FlowCell filter functions."""

    BY_INTERNAL_ID: callable = filter_illumina_flow_cell_by_internal_id



from pathlib import Path
from typing import Any, Type

from cg.constants.constants import FileFormat
from cg.constants.pacbio import PacBioDirsAndFiles
from cg.io.controller import ReadFile
from cg.services.run_devices.pacbio.metrics_parser.models import (
    BaseMetrics,
    ControlMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.utils.files import get_file_with_pattern_from_list


def _get_data_model_from_pattern(pattern: str) -> Type[BaseMetrics]:
    """Return the data model based on the pattern."""
    pattern_to_model = {
        PacBioDirsAndFiles.CCS_REPORT_SUFFIX: ReadMetrics,
        PacBioDirsAndFiles.CONTROL_REPORT: ControlMetrics,
        PacBioDirsAndFiles.LOADING_REPORT: ProductivityMetrics,
        PacBioDirsAndFiles.RAW_DATA_REPORT: PolymeraseMetrics,
        PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT: SmrtlinkDatasetsMetrics,
    }
    return pattern_to_model.get(pattern)


def _parse_report_to_model(report_file: Path, data_model: Type[BaseMetrics]) -> BaseMetrics:
    """Parse the metrics report to a data model."""
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report_file)
    if data_model == SmrtlinkDatasetsMetrics:
        return data_model.model_validate(parsed_json[0], from_attributes=True)
    metrics: list[dict[str, Any]] = parsed_json.get("attributes")
    data: dict = {report_field["id"]: report_field["value"] for report_field in metrics}
    return data_model.model_validate(data, from_attributes=True)


def get_parsed_metrics_from_file_name(metrics_files: list[Path], file_name: str) -> BaseMetrics:
    report_file: Path = get_file_with_pattern_from_list(files=metrics_files, pattern=file_name)
    data_model: Type[BaseMetrics] = _get_data_model_from_pattern(pattern=file_name)
    return _parse_report_to_model(report_file=report_file, data_model=data_model)



from typing import Any

from pydantic.v1 import BaseModel
from typing_extensions import Literal

from cg.apps.lims import LimsAPI

from cg.constants.lims import LimsArtifactTypes, LimsProcess
from cg.exc import LimsDataError


class LimsProject(BaseModel):
    id: str = "1"


class MockReagentType:
    """Mock class for reagent type."""

    def __init__(self, label: str, sequence: str):
        self.label: str = label
        self.sequence: str = sequence


class LimsUDF(BaseModel):
    control: str | None = None
    customer: str = None

    def get(self, argument: str) -> str:
        if argument == "Control":
            return self.control
        if argument == "customer":
            return self.customer


class LimsSample(BaseModel):
    id: str
    name: str = None
    customer: str = None
    sex: Literal["male", "female", "unknown"] = None
    father: str = None
    mother: str = None
    status: Literal["affected", "unaffected", "unknown"] = None
    application: str = None
    application_version: str = None
    family: str = None
    panels: list[str] = None
    comment: str = None
    project: LimsProject = LimsProject()
    received: str = None
    source: str = None
    priority: str = None
    udfs: LimsUDF = LimsUDF()


class LimsArtifactObject(BaseModel):
    parent_process: LimsProcess = LimsProcess.COVID_POOLING_STEP
    type: LimsArtifactTypes = LimsArtifactTypes.ANALYTE
    samples: list[LimsSample] = []


class MockLimsAPI(LimsAPI):
    """Mock LIMS API for testing."""

    def __init__(self, config: dict = None, samples: list[dict] = None):
        if samples is None:
            samples = []
        self.config = config
        self.baseuri = "https://clinical-lims-mock.scilifelab.se"
        self._received_at = None
        self.sample_vars = {}
        self._samples = samples
        self._prep_method = (
            "CG002 - End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free "
            ""
            "DNA)"
        )
        self._sequencing_method = "CG002 - Cluster Generation (HiSeq X)"
        self._delivery_method = "CG002 - Delivery"
        self._source = "cell-free DNA"
        self.artifacts: dict[str, list[LimsArtifactObject]] = {}

    def set_prep_method(self, method: str = "1337:00 Test prep method"):
        """Mock function"""
        self._prep_method = method

    def set_sequencing_method(self, method: str = "1338:00 Test sequencing method"):
        """Mock function"""
        self._prep_method = method

    def sample(self, lims_id: str) -> dict:
        return next((sample for sample in self._samples if sample["id"] == lims_id), {})

    def add_sample(self, internal_id: str):
        self.sample_vars[internal_id] = {}

    def add_capture_kit(self, internal_id: str, capture_kit):
        if internal_id not in self.sample_vars:
            self.add_sample(internal_id)
        self.sample_vars[internal_id]["capture_kit"] = capture_kit

    def add_artifact_for_sample(
        self,
        sample_id: str,
        samples: list[LimsSample] = None,
    ):
        if sample_id in self.artifacts:
            self.artifacts[sample_id].append(LimsArtifactObject(samples=samples))
        else:
            self.artifacts[sample_id] = [LimsArtifactObject(samples=samples)]

    def get_latest_artifact_for_sample(
        self, process_type: LimsProcess, artifact_type: LimsArtifactTypes, sample_internal_id: str
    ) -> LimsArtifactObject:
        return self.artifacts[sample_internal_id][0]

    def get_internal_negative_control_id_from_sample_in_pool(
        self, sample_internal_id: str, pooling_step: LimsProcess
    ) -> str:
        """Retrieve from lims the sample_id for the internal negative control sample present in the same pool as the given sample."""
        artifact: LimsArtifactObject = self.get_latest_artifact_for_sample(
            process_type=pooling_step,
            artifact_type=LimsArtifactTypes.ANALYTE,
            sample_internal_id=sample_internal_id,
        )
        samples = artifact.samples

        negative_controls: list = self._get_negative_controls_from_list(samples=samples)

        if len(negative_controls) > 1:
            sample_ids = [sample.id for sample in negative_controls]
            raise LimsDataError(
                f"Several internal negative control samples found: {' '.join(sample_ids)}"
            )
        return negative_controls[0].id

    @staticmethod
    def _get_negative_controls_from_list(samples: list[LimsSample]) -> list[LimsSample]:
        """Filter and return a list of internal negative controls from a given sample list."""
        negative_controls = []
        for sample in samples:
            if sample.udfs.control == "negative" and sample.udfs.customer == "cust000":
                negative_controls.append(sample)
        return negative_controls

    def capture_kit(self, lims_id: str):
        if lims_id in self.sample_vars:
            return self.sample_vars[lims_id].get("capture_kit")
        return None

    def get_prep_method(self, lims_id: str) -> str:
        return self._prep_method

    def get_sequencing_method(self, lims_id: str) -> str:
        return self._sequencing_method

    def get_delivery_method(self, lims_id: str) -> str:
        return self._delivery_method

    def get_source(self, lims_id: str) -> str | None:
        return self._source

    def get_sample_comment(self, sample_id: str) -> str:
        lims_sample: dict[str, Any] = self.sample(sample_id)
        comment = None
        if lims_sample:
            comment: str = lims_sample.get("comment")
        return comment

    def get_sample_project(self, sample_id: str) -> str | None:
        lims_sample: dict[str, Any] = self.sample(sample_id)
        project_id = None
        if lims_sample:
            project_id: str = lims_sample.get("project").get("id")
        return project_id

    def update_sample(
        self, lims_id: str, sex=None, target_reads: int = None, name: str = None, **kwargs
    ):
        pass  # This is completely mocked out

    def set_samples(self, samples):
        self._samples = samples

    def cache(self):
        pass  # This is completely mocked out

    def get_received_date(self, lims_id: str):
        received_date = None
        for sample in self._samples:
            if sample.internal_id == lims_id:
                received_date = sample.received_at
        return received_date

    def get_sample_rin(self, sample_id: str) -> float:
        """Mock return sample RIN value."""
        return 10.0

    def get_latest_rna_input_amount(self, sample_id: str) -> float:
        """Mock return input amount used in the latest preparation of an RNA sample."""
        return 300.0

    def get_sample_dv200(self, sample_id: str) -> float:
        """Mock return sample's percentage of RNA fragments greater than 200 nucleotides."""
        return 75.0

    def has_sample_passed_initial_qc(self, sample_id: str) -> bool:
        """Mock return of the sample initial QC flag."""
        return True



"""Test the move delivery files service."""

import pytest

from cg.services.deliver_files.delivery_file_fetcher_service.models import DeliveryFiles
from cg.services.deliver_files.delivery_file_mover_service.delivery_file_mover import (
    DeliveryFilesMover,
)


@pytest.mark.parametrize(
    "expected_moved_delivery_files,delivery_files",
    [
        ("expected_moved_fastq_delivery_files", "expected_fastq_delivery_files"),
        ("expected_moved_analysis_delivery_files", "expected_analysis_delivery_files"),
    ],
)
def test_move_files(
    expected_moved_delivery_files: DeliveryFiles,
    delivery_files: DeliveryFiles,
    tmp_path,
    request,
):
    # GIVEN a delivery files to move
    expected_moved_delivery_files: DeliveryFiles = request.getfixturevalue(
        expected_moved_delivery_files
    )
    delivery_files: DeliveryFiles = request.getfixturevalue(delivery_files)

    # WHEN moving the delivery files
    move_files_service = DeliveryFilesMover()
    moved_delivery_files: DeliveryFiles = move_files_service.move_files(
        delivery_files=delivery_files, delivery_base_path=tmp_path
    )

    # THEN assert that the delivery files are moved
    assert moved_delivery_files == expected_moved_delivery_files
    if case_files := moved_delivery_files.case_files:
        for case_file in case_files:
            assert case_file.file_path.exists()
    for sample_file in moved_delivery_files.sample_files:
        assert sample_file.file_path.exists()



"""Test Balsamic observations API."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockFixture

from cg.constants.constants import CancerAnalysisType
from cg.exc import LoqusdbDuplicateRecordError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.observations.input_files import BalsamicObservationsInputFiles
from cg.store.models import Case


def test_is_analysis_type_eligible_for_observations_upload(
    case_id: str, balsamic_observations_api: BalsamicObservationsAPI, mocker: MockFixture
):
    """Test if the analysis type is eligible for observation uploads."""

    # GIVEN a case ID and a Balsamic observations API

    # GIVEN a case with tumor samples
    mocker.patch.object(BalsamicAnalysisAPI, "is_analysis_normal_only", return_value=False)

    # WHEN checking analysis type eligibility for a case
    is_analysis_type_eligible_for_observations_upload: bool = (
        balsamic_observations_api.is_analysis_type_eligible_for_observations_upload(case_id)
    )

    # THEN the analysis type should be eligible for observation uploads
    assert is_analysis_type_eligible_for_observations_upload


def test_is_analysis_type_not_eligible_for_observations_upload(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test if the analysis type is not eligible for observation uploads."""

    # GIVEN a case ID and a Balsamic observations API

    # GIVEN a case without tumor samples (normal-only analysis)
    mocker.patch.object(BalsamicAnalysisAPI, "is_analysis_normal_only", return_value=True)

    # WHEN checking analysis type eligibility for a case
    is_analysis_type_eligible_for_observations_upload: bool = (
        balsamic_observations_api.is_analysis_type_eligible_for_observations_upload(case_id)
    )

    # THEN the analysis type should not be eligible for observation uploads
    assert not is_analysis_type_eligible_for_observations_upload
    assert f"Normal only analysis {case_id} is not supported for Loqusdb uploads" in caplog.text


def test_is_case_eligible_for_observations_upload(
    case_id: str, balsamic_observations_api: BalsamicObservationsAPI, mocker: MockFixture
):
    """Test whether a case is eligible for Balsamic observation uploads."""

    # GIVEN a case and a Balsamic observations API
    case: Case = balsamic_observations_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN an eligible sequencing method and a case with tumor samples
    mocker.patch.object(
        BalsamicAnalysisAPI, "get_data_analysis_type", return_value=CancerAnalysisType.TUMOR_WGS
    )
    mocker.patch.object(BalsamicAnalysisAPI, "is_analysis_normal_only", return_value=False)

    # WHEN checking the upload eligibility for a case
    is_case_eligible_for_observations_upload: bool = (
        balsamic_observations_api.is_case_eligible_for_observations_upload(case)
    )

    # THEN the case should be eligible for observation uploads
    assert is_case_eligible_for_observations_upload


def test_is_case_not_eligible_for_observations_upload(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test whether a case is eligible for Balsamic observation uploads."""

    # GIVEN a case and a Balsamic observations API
    case: Case = balsamic_observations_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a case with tumor sample and an invalid sequencing type
    mocker.patch.object(BalsamicAnalysisAPI, "is_analysis_normal_only", return_value=False)
    mocker.patch.object(
        BalsamicAnalysisAPI, "get_data_analysis_type", return_value=CancerAnalysisType.TUMOR_PANEL
    )

    # WHEN checking the upload eligibility for a case
    is_case_eligible_for_observations_upload: bool = (
        balsamic_observations_api.is_case_eligible_for_observations_upload(case)
    )

    # THEN the case should not be eligible for observation uploads
    assert not is_case_eligible_for_observations_upload


def test_load_observations(
    case_id: str,
    loqusdb_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    balsamic_observations_input_files: BalsamicObservationsInputFiles,
    number_of_loaded_variants: int,
    caplog: LogCaptureFixture,
    mocker: MockFixture,
):
    """Test loading of Balsamic case observations."""
    caplog.set_level(logging.DEBUG)

    # GIVEN an observations API and a list of input files for upload
    case: Case = balsamic_observations_api.store.get_case_by_internal_id(case_id)
    mocker.patch.object(
        BalsamicObservationsAPI,
        "get_observations_input_files",
        return_value=balsamic_observations_input_files,
    )

    # GIVEN an observations API mocked scenario
    mocker.patch.object(BalsamicObservationsAPI, "is_duplicate", return_value=False)

    # WHEN loading the case to Loqusdb
    balsamic_observations_api.load_observations(case)

    # THEN the observations should be loaded successfully
    assert f"Uploaded {number_of_loaded_variants} variants to Loqusdb" in caplog.text


def test_load_duplicated_observations(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    caplog: LogCaptureFixture,
    mocker: MockFixture,
):
    """Test raise of a duplicate exception when loading Balsamic case observations."""
    caplog.set_level(logging.DEBUG)

    # GIVEN an observations API and a Balsamic case
    case: Case = balsamic_observations_api.store.get_case_by_internal_id(case_id)

    # GIVEN a duplicate case in Loqusdb
    mocker.patch.object(BalsamicObservationsAPI, "is_duplicate", return_value=True)

    # WHEN loading the case to Loqusdb
    with pytest.raises(LoqusdbDuplicateRecordError):
        balsamic_observations_api.load_observations(case)

    # THEN the observations upload should be aborted
    assert f"Case {case_id} has already been uploaded to Loqusdb" in caplog.text


def test_load_cancer_observations(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    balsamic_observations_input_files: BalsamicObservationsInputFiles,
    number_of_loaded_variants: int,
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """Test loading of cancer case observations for Balsamic."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a Balsamic observations API, a list of input files, and a cancer case
    case: Case = balsamic_observations_api.store.get_case_by_internal_id(case_id)

    # GIVEN an observations API mocked scenario
    mocker.patch.object(
        BalsamicAnalysisAPI,
        "get_data_analysis_type",
        return_value=CancerAnalysisType.TUMOR_NORMAL_WGS,
    )

    # WHEN loading the case to a somatic Loqusdb instance
    balsamic_observations_api.load_cancer_observations(
        case=case,
        input_files=balsamic_observations_input_files,
        loqusdb_api=balsamic_observations_api.loqusdb_somatic_api,
    )

    # THEN the observations should be loaded successfully
    assert f"Uploaded {number_of_loaded_variants} variants to Loqusdb" in caplog.text



"""Module for Flask-Admin views"""

from gettext import gettext

from flask import flash, redirect, request, session, url_for
from flask_admin.actions import action
from flask_admin.contrib.sqla import ModelView
from flask_dance.contrib.google import google
from markupsafe import Markup

from cg.constants.constants import NG_UL_SUFFIX, CaseActions, DataDelivery, Workflow
from cg.server.ext import db, sample_service
from cg.utils.flask.enum import SelectEnumField


class BaseView(ModelView):
    """Base for the specific views."""

    def is_accessible(self):
        user = db.get_user_by_email(email=session.get("user_email"))
        return bool(google.authorized and user and user.is_admin)

    def inaccessible_callback(self, name, **kwargs):
        # redirect to login page if user doesn't have access
        return redirect(url_for("google.login", next=request.url))


def view_priority(unused1, unused2, model, unused3):
    """column formatter for priority"""
    del unused1, unused2, unused3
    return Markup("%s" % model.priority.name) if model else ""


def view_flow_cell_internal_id(unused1, unused2, model, unused3):
    """column formatter for priority"""
    del unused1, unused2, unused3
    return Markup("%s" % model.device.internal_id)


def view_flow_cell_model(unused1, unused2, model, unused3):
    """column formatter for priority"""
    del unused1, unused2, unused3
    return Markup("%s" % model.device.model)


def view_case_sample_link(unused1, unused2, model, unused3):
    """column formatter to open the case-sample view"""

    del unused1, unused2, unused3

    return Markup(
        "<a href='%s'>%s</a>"
        % (url_for("casesample.index_view", search=model.internal_id), model.internal_id)
    )


def is_external_application(unused1, unused2, model, unused3):
    """column formatter to open this view"""
    del unused1, unused2, unused3
    return model.application_version.application.is_external if model.application_version else ""


def view_sample_concentration_minimum(unused1, unused2, model, unused3):
    """Column formatter to append unit"""
    del unused1, unused2, unused3
    return (
        str(model.sample_concentration_minimum) + NG_UL_SUFFIX
        if model.sample_concentration_minimum
        else None
    )


def view_sample_concentration_maximum(unused1, unused2, model, unused3):
    """Column formatter to append unit"""
    del unused1, unused2, unused3
    return (
        str(model.sample_concentration_maximum) + NG_UL_SUFFIX
        if model.sample_concentration_maximum
        else None
    )


def view_sample_concentration_minimum_cfdna(unused1, unused2, model, unused3):
    """Column formatter to append unit"""
    del unused1, unused2, unused3
    return (
        str(model.sample_concentration_minimum_cfdna) + NG_UL_SUFFIX
        if model.sample_concentration_minimum_cfdna
        else None
    )


def view_sample_concentration_maximum_cfdna(unused1, unused2, model, unused3):
    """Column formatter to append unit"""
    del unused1, unused2, unused3
    return (
        str(model.sample_concentration_maximum_cfdna) + NG_UL_SUFFIX
        if model.sample_concentration_maximum_cfdna
        else None
    )


class ApplicationView(BaseView):
    """Admin view for Model.Application"""

    column_editable_list = [
        "description",
        "is_accredited",
        "target_reads",
        "percent_reads_guaranteed",
        "comment",
        "prep_category",
        "sequencing_depth",
        "min_sequencing_depth",
        "is_external",
        "turnaround_time",
        "sample_concentration",
        "sample_concentration_minimum",
        "sample_concentration_maximum",
        "sample_concentration_minimum_cfdna",
        "sample_concentration_maximum_cfdna",
        "priority_processing",
        "is_archived",
    ]
    column_exclude_list = [
        "minimum_order",
        "sample_amount",
        "sample_volume",
        "details",
        "limitations",
        "created_at",
        "updated_at",
        "category",
    ]
    column_formatters = {
        "sample_concentration_minimum": view_sample_concentration_minimum,
        "sample_concentration_maximum": view_sample_concentration_maximum,
        "sample_concentration_minimum_cfdna": view_sample_concentration_minimum_cfdna,
        "sample_concentration_maximum_cfdna": view_sample_concentration_maximum_cfdna,
    }
    column_filters = ["prep_category", "is_accredited"]
    column_searchable_list = ["tag", "prep_category"]
    form_excluded_columns = ["category", "versions"]

    @staticmethod
    def view_application_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (
                    url_for("application.index_view", search=model.application.tag),
                    model.application.tag,
                )
            )
            if model.application
            else ""
        )


class ApplicationVersionView(BaseView):
    """Admin view for Model.ApplicationVersion"""

    column_default_sort = ("valid_from", True)
    column_editable_list = [
        "valid_from",
        "price_standard",
        "price_priority",
        "price_express",
        "price_clinical_trials",
        "price_research",
    ]
    column_list = (
        "application",
        "version",
        "valid_from",
        "price_standard",
        "price_priority",
        "price_express",
        "price_clinical_trials",
        "price_research",
        "comment",
    )
    column_exclude_list = ["created_at", "updated_at"]
    column_filters = ["version", "application.tag"]
    column_formatters = {"application": ApplicationView.view_application_link}
    column_searchable_list = ["application.tag"]
    edit_modal = True
    create_modal = True
    form_excluded_columns = ["samples", "pools", "microbial_samples"]


class ApplicationLimitationsView(BaseView):
    """Admin view for Model.ApplicationLimitations."""

    column_list = (
        "application",
        "workflow",
        "limitations",
        "comment",
        "created_at",
        "updated_at",
    )
    column_formatters = {"application": ApplicationView.view_application_link}
    column_filters = ["application.tag", "workflow"]
    column_searchable_list = ["application.tag"]
    column_editable_list = ["comment"]
    form_excluded_columns = ["created_at", "updated_at"]
    form_extra_fields = {"workflow": SelectEnumField(enum_class=Workflow)}
    create_modal = True
    edit_modal = True


class BedView(BaseView):
    """Admin view for Model.Bed"""

    column_default_sort = "name"
    column_editable_list = ["comment"]
    column_exclude_list = ["created_at"]
    column_filters = []
    column_searchable_list = ["name"]
    form_excluded_columns = ["created_at", "updated_at"]

    @staticmethod
    def view_bed_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (url_for("bed.index_view", search=model.bed.name), model.bed.name)
            )
            if model.bed
            else ""
        )


class BedVersionView(BaseView):
    """Admin view for Model.BedVersion"""

    column_default_sort = ("updated_at", True)
    column_editable_list = ["shortname", "filename", "comment", "designer", "checksum"]
    column_exclude_list = ["created_at"]
    form_excluded_columns = ["created_at", "updated_at", "samples"]
    column_filters = []
    column_formatters = {"bed": BedView.view_bed_link}
    column_searchable_list = ["bed.name"]
    edit_modal = True


class CustomerView(BaseView):
    """Admin view for Model.Customer"""

    column_editable_list = [
        "collaborations",
        "comment",
        "delivery_contact",
        "lab_contact",
        "loqus_upload",
        "primary_contact",
        "priority",
        "return_samples",
        "scout_access",
    ]
    column_list = [
        "internal_id",
        "name",
        "data_archive_location",
        "comment",
        "primary_contact",
        "delivery_contact",
        "lab_contact",
        "priority",
        "project_account_KI",
        "project_account_kth",
        "return_samples",
        "scout_access",
    ]
    column_filters = ["priority", "scout_access", "data_archive_location"]
    column_searchable_list = ["internal_id", "name"]
    form_excluded_columns = ["families", "samples", "pools", "orders", "invoices"]


class CollaborationView(BaseView):
    """Admin view for Model.CustomerGroup"""

    column_editable_list = ["name"]
    column_filters = []
    column_hide_backrefs = False
    column_list = ("internal_id", "name", "customers")
    column_searchable_list = ["internal_id", "name"]


class CaseView(BaseView):
    """Admin view for Model.Case"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["action", "comment"]
    column_exclude_list = ["created_at", "_cohorts", "synopsis"]
    column_filters = [
        "customer.internal_id",
        "priority",
        "action",
        "data_analysis",
        "data_delivery",
        "tickets",
    ]
    column_formatters = {
        "internal_id": view_case_sample_link,
        "priority": view_priority,
    }
    column_searchable_list = [
        "internal_id",
        "name",
        "customer.internal_id",
        "tickets",
    ]
    form_excluded_columns = [
        "analyses",
        "_cohorts",
        "links",
        "synopsis",
    ]
    form_extra_fields = {
        "data_analysis": SelectEnumField(enum_class=Workflow),
        "data_delivery": SelectEnumField(enum_class=DataDelivery),
    }

    @staticmethod
    def view_case_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        markup = ""
        if model.case:
            markup += Markup(
                " <a href='%s'>%s</a>"
                % (url_for("case.index_view", search=model.case.internal_id), model.case)
            )

        return markup

    @action(
        "set_hold",
        "Set action to hold",
        "Are you sure you want to set the action for selected families to hold?",
    )
    def action_set_hold(self, ids: list[str]):
        self.set_action_for_cases(action=CaseActions.HOLD, case_entry_ids=ids)

    @action(
        "set_empty",
        "Set action to Empty",
        "Are you sure you want to set the action for selected families to Empty?",
    )
    def action_set_empty(self, ids: list[str]):
        self.set_action_for_cases(action=None, case_entry_ids=ids)

    def set_action_for_cases(self, action: CaseActions | None, case_entry_ids: list[str]):
        try:
            for entry_id in case_entry_ids:
                case = db.get_case_by_entry_id(entry_id=entry_id)
                if case:
                    case.action = action

            db.session.commit()

            num_families = len(case_entry_ids)
            action_message = (
                f"Families were set to {action}."
                if num_families == 1
                else f"{num_families} families were set to {action}."
            )
            flash(action_message)

        except Exception as ex:
            if not self.handle_view_exception(ex):
                raise

            flash(gettext(f"Failed to set case action. {str(ex)}"))


class InvoiceView(BaseView):
    """Admin view for Model.Invoice"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["comment"]
    column_list = (
        "id",
        "customer",
        "created_at",
        "updated_at",
        "invoiced_at",
        "comment",
        "discount",
        "price",
    )
    column_searchable_list = ["customer.internal_id", "customer.name", "id"]

    @staticmethod
    def view_invoice_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (
                    url_for("invoice.index_view", search=model.invoice.id),
                    (
                        model.invoice.invoiced_at.date()
                        if model.invoice.invoiced_at
                        else "In progress"
                    ),
                )
            )
            if model.invoice
            else ""
        )


class AnalysisView(BaseView):
    """Admin view for Model.Analysis"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["is_primary", "comment"]
    column_filters = ["workflow", "workflow_version", "is_primary"]
    column_formatters = {"case": CaseView.view_case_link}
    column_searchable_list = [
        "case.internal_id",
        "case.name",
    ]
    form_extra_fields = {"workflow": SelectEnumField(enum_class=Workflow)}


class IlluminaFlowCellView(BaseView):
    """Admin view for Model.IlluminaSequencingRun"""

    column_list = (
        "internal_id",
        "model",
        "sequencer_type",
        "sequencer_name",
        "data_availability",
        "has_backup",
        "total_reads",
        "total_undetermined_reads",
        "percent_undetermined_reads",
        "percent_q30",
        "mean_quality_score",
        "total_yield",
        "yield_q30",
        "cycles",
        "demultiplexing_software",
        "demultiplexing_software_version",
        "sequencing_started_at",
        "sequencing_completed_at",
        "demultiplexing_started_at",
        "demultiplexing_completed_at",
        "archived_at",
    )
    column_formatters = {"internal_id": view_flow_cell_internal_id, "model": view_flow_cell_model}
    column_default_sort = ("sequencing_completed_at", True)
    column_filters = ["sequencer_type", "sequencer_name", "data_availability"]
    column_editable_list = ["data_availability"]
    column_searchable_list = ["sequencer_type", "sequencer_name", "device.internal_id"]
    column_sortable_list = [
        ("internal_id", "device.internal_id"),
        "sequencer_type",
        "sequencer_name",
        "data_availability",
        "has_backup",
        "sequencing_started_at",
        "sequencing_completed_at",
        "demultiplexing_started_at",
        "demultiplexing_completed_at",
        "archived_at",
    ]

    @staticmethod
    def view_flow_cell_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (
                    url_for(
                        "illuminasequencingrun.index_view",
                        search=model.instrument_run.device.internal_id,
                    ),
                    model.instrument_run.device.internal_id,
                )
            )
            if model.instrument_run.device
            else ""
        )


class OrganismView(BaseView):
    """Admin view for Model.Organism"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["internal_id", "name", "reference_genome", "comment"]
    column_searchable_list = ["internal_id", "name", "reference_genome"]


class OrderView(BaseView):
    """Admin view for Model.Order"""

    column_default_sort = ("order_date", True)
    column_editable_list = ["is_delivered"]
    column_searchable_list = ["id", "ticket_id"]
    column_display_pk = True
    create_modal = True
    edit_modal = True


class PanelView(BaseView):
    """Admin view for Model.Panel"""

    column_editable_list = ["current_version", "name"]
    column_filters = ["customer.internal_id"]
    column_searchable_list = ["customer.internal_id", "name", "abbrev"]
    create_modal = True
    edit_modal = True


class PoolView(BaseView):
    """Admin view for Model.Pool"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["ticket"]
    column_filters = ["customer.internal_id", "application_version.application"]
    column_formatters = {"invoice": InvoiceView.view_invoice_link}
    column_searchable_list = ["name", "order", "ticket", "customer.internal_id"]


class SampleView(BaseView):
    """Admin view for Model.Sample"""

    column_exclude_list = [
        "age_at_sampling",
        "_phenotype_groups",
        "_phenotype_terms",
    ]
    column_default_sort = ("created_at", True)
    column_editable_list = [
        "comment",
        "downsampled_to",
        "is_tumour",
        "last_sequenced_at",
        "sex",
    ]
    column_filters = [
        "customer.internal_id",
        "priority",
        "sex",
        "application_version.application",
        "capture_kit",
    ]
    column_formatters = {
        "is_external": is_external_application,
        "internal_id": view_case_sample_link,
        "invoice": InvoiceView.view_invoice_link,
        "priority": view_priority,
    }
    column_searchable_list = [
        "internal_id",
        "name",
        "subject_id",
        "customer.internal_id",
        "original_ticket",
    ]
    form_excluded_columns = [
        "age_at_sampling",
        "deliveries",
        "father_links",
        "flowcells",
        "invoice",
        "_phenotype_groups",
        "_phenotype_terms",
        "links",
        "mother_links",
    ]

    @staticmethod
    def view_sample_link(unused1, unused2, model, unused3):
        """column formatter to open this view"""
        del unused1, unused2, unused3
        return (
            Markup(
                "<a href='%s'>%s</a>"
                % (url_for("sample.index_view", search=model.sample.internal_id), model.sample)
            )
            if model.sample
            else ""
        )

    @action(
        "cancel_samples",
        "Cancel samples",
        "Are you sure you want to cancel the selected samples?",
    )
    def cancel_samples(self, entry_ids: list[str]) -> None:
        user_email: str | None = session.get("user_email")
        message: str = sample_service.cancel_samples(
            sample_ids=entry_ids,
            user_email=user_email,
        )
        flash(message)


class DeliveryView(BaseView):
    """Admin view for Model.Delivery"""

    column_default_sort = ("id", True)
    column_filters = ["sample.internal_id"]
    column_formatters = {"sample": SampleView.view_sample_link}
    column_searchable_list = ["sample.internal_id"]
    create_modal = True
    edit_modal = True


class CaseSampleView(BaseView):
    """Admin view for Model.caseSample"""

    column_default_sort = ("created_at", True)
    column_editable_list = ["status"]
    column_filters = ["status"]
    column_formatters = {
        "case": CaseView.view_case_link,
        "sample": SampleView.view_sample_link,
    }
    column_searchable_list = ["case.internal_id", "case.name", "sample.internal_id"]
    create_modal = True
    edit_modal = True


class UserView(BaseView):
    """Admin view for Model.User"""

    column_default_sort = "name"
    column_editable_list = ["order_portal_login"]
    column_filters = ["is_admin", "order_portal_login", "customers"]
    column_hide_backrefs = False
    column_list = ("name", "email", "is_admin", "order_portal_login", "customers")
    column_searchable_list = ["name", "email"]
    create_modal = True
    edit_modal = True


class IlluminaSampleSequencingMetricsView(BaseView):
    column_list = [
        "flow_cell",
        "sample",
        "flow_cell_lane",
        "total_reads_in_lane",
        "base_passing_q30_percent",
        "base_mean_quality_score",
        "yield_",
        "yield_q30",
        "created_at",
    ]
    column_formatters = {
        "flow_cell": IlluminaFlowCellView.view_flow_cell_link,
        "sample": SampleView.view_sample_link,
    }
    column_searchable_list = ["sample.internal_id", "instrument_run.device.internal_id"]



"""Post request model for JanusAPI."""

from pydantic import BaseModel

from cg.constants import Workflow


class WorkflowInfo(BaseModel):
    workflow: Workflow
    version: str


class FilePathAndTag(BaseModel):
    """Model for the file path and its associated tag."""

    file_path: str
    tag: str


class CreateQCMetricsRequest(BaseModel):
    """Model for the qc collection request."""

    case_id: str
    sample_ids: list[str]
    files: list[FilePathAndTag]
    workflow_info: WorkflowInfo



from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Workflow
from cg.services.deliver_files.delivery_file_tag_fetcher_service.sample_and_case_delivery_tags_fetcher import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.delivery_file_fetcher_service.error_handling import (
    handle_missing_bundle_errors,
)
from cg.services.deliver_files.delivery_file_fetcher_service.delivery_file_fetcher_service import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.delivery_file_fetcher_service.models import (
    SampleFile,
    DeliveryFiles,
    DeliveryMetaData,
)
from cg.store.models import Case
from cg.store.store import Store
from housekeeper.store.models import File


class FastqDeliveryFileFetcher(FetchDeliveryFilesService):
    """
    Fetch the fastq files for a case from Housekeeper.
    """

    def __init__(
        self,
        status_db: Store,
        hk_api: HousekeeperAPI,
        tags_fetcher: SampleAndCaseDeliveryTagsFetcher,
    ):
        self.status_db = status_db
        self.hk_api = hk_api
        self.tags_fetcher = tags_fetcher

    def get_files_to_deliver(self, case_id: str) -> DeliveryFiles:
        """Return a list of FASTQ files to be delivered for a case and its samples."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_ids: list[str] = case.sample_ids
        fastq_files: list[SampleFile] = []
        for sample_id in sample_ids:
            if sample_fastq_files := self._get_fastq_files_for_sample(
                case_id=case_id, sample_id=sample_id
            ):
                fastq_files.extend(sample_fastq_files)

        delivery_data = DeliveryMetaData(
            customer_internal_id=case.customer.internal_id, ticket_id=case.latest_ticket
        )
        return DeliveryFiles(
            delivery_data=delivery_data,
            case_files=None,
            sample_files=fastq_files,
        )

    @handle_missing_bundle_errors
    def _get_fastq_files_for_sample(self, case_id: str, sample_id: str) -> list[SampleFile] | None:
        """Get the FASTQ files for a sample."""
        fastq_tags: list[set[str]] = self.tags_fetcher.fetch_tags(Workflow.FASTQ).sample_tags
        fastq_files: list[File] = self.hk_api.get_files_from_latest_version_containing_tags(
            bundle_name=sample_id, tags=fastq_tags
        )
        sample_name: str = self.status_db.get_sample_by_internal_id(sample_id).name
        return [
            SampleFile(
                case_id=case_id,
                sample_id=sample_id,
                sample_name=sample_name,
                file_path=fastq_file.full_path,
            )
            for fastq_file in fastq_files
        ]



from pathlib import Path
from unittest import mock

import pytest

from cg.services.run_devices.exc import PostProcessingRunFileManagerError
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager


def test_get_files_to_parse(
    expected_pac_bio_run_data: PacBioRunData,
    pac_bio_report_files_to_parse: list[Path],
    pac_bio_run_file_manager: PacBioRunFileManager,
):
    # GIVEN a run data object and a PacBio run file manager

    # WHEN getting the files to parse
    files: list[Path] = pac_bio_run_file_manager.get_files_to_parse(expected_pac_bio_run_data)

    # THEN the correct files are returned
    assert files == pac_bio_report_files_to_parse


def test_get_files_to_store(
    expected_pac_bio_run_data: PacBioRunData,
    pac_bio_run_file_manager: PacBioRunFileManager,
    pac_bio_report_files_to_parse: list[Path],
    pac_bio_hifi_read_file: Path,
):
    # GIVEN a run data object and a PacBio file manager

    # WHEN getting the files to store
    files: list[Path] = pac_bio_run_file_manager.get_files_to_store(expected_pac_bio_run_data)

    # THEN the correct files are returned
    full_list: list[Path] = pac_bio_report_files_to_parse + [pac_bio_hifi_read_file]
    assert set(files) == set(full_list)


def test_get_files_to_store_error(
    expected_pac_bio_run_data: PacBioRunData,
    pac_bio_run_file_manager: PacBioRunFileManager,
):
    # GIVEN a run data object

    # GIVEN a PacBio run file manager that can't find the HiFi read file
    with mock.patch.object(
        pac_bio_run_file_manager,
        attribute="_get_hifi_read_files",
        side_effect=FileNotFoundError,
    ):
        # WHEN getting the files to store

        # THEN an PostProcessingRunFileManagerError is raised
        with pytest.raises(PostProcessingRunFileManagerError):
            pac_bio_run_file_manager.get_files_to_store(expected_pac_bio_run_data)


def test_get_files_to_parse_error(
    expected_pac_bio_run_data: PacBioRunData,
    pac_bio_run_file_manager: PacBioRunFileManager,
):
    # GIVEN a run data object

    # GIVEN a PacBio run file manager that can't find the CCS report file
    with mock.patch.object(
        pac_bio_run_file_manager,
        attribute="_get_ccs_report_file",
        side_effect=FileNotFoundError,
    ):
        # WHEN getting the files to store

        # THEN an PostProcessingRunFileManagerError is raised
        with pytest.raises(PostProcessingRunFileManagerError):
            pac_bio_run_file_manager.get_files_to_parse(expected_pac_bio_run_data)



"""Module for PacBio fixtures returning strings."""

import pytest


@pytest.fixture
def pac_bio_smrt_cell_name() -> str:
    return "1_A01"


@pytest.fixture
def pac_bio_test_run_name() -> str:
    """Return the name of a PacBio SMRT cell."""
    return "r84202_20240522_133539"


@pytest.fixture
def pac_bio_sequencing_run_name(pac_bio_test_run_name: str, pac_bio_smrt_cell_name: str) -> str:
    """Return the name of a PacBio SMRT cell."""
    return f"{pac_bio_test_run_name}/{pac_bio_smrt_cell_name})"


@pytest.fixture
def pac_bio_1_a01_cell_full_name() -> str:
    """Return the full name of a PacBio SMRT cell."""
    return "m84202_240522_135641_s1"


@pytest.fixture
def smrt_cell_internal_id() -> str:
    return "EA094834"


@pytest.fixture
def pac_bio_sample_internal_id() -> str:
    return "1247014000119"



import logging

from cg.constants import DEFAULT_CAPTURE_KIT, Workflow
from cg.constants.constants import AnalysisType
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.pedigree import Pedigree
from cg.constants.scout import MIP_CASE_TAGS
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.mip.mip_analysis import MipAnalysis
from cg.store.models import Case, CaseSample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MipDNAAnalysisAPI(MipAnalysisAPI):
    def __init__(self, config: CGConfig, workflow: Workflow = Workflow.MIP_DNA):
        super().__init__(config, workflow)

    @property
    def root(self) -> str:
        return self.config.mip_rd_dna.root

    @property
    def conda_binary(self) -> str:
        return self.config.mip_rd_dna.conda_binary

    @property
    def conda_env(self) -> str:
        return self.config.mip_rd_dna.conda_env

    @property
    def mip_workflow(self) -> str:
        return self.config.mip_rd_dna.workflow

    @property
    def script(self) -> str:
        return self.config.mip_rd_dna.script

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=f"{self.script} {self.mip_workflow}",
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                config=self.config.mip_rd_dna.mip_config,
                environment=self.conda_env,
            )
        return self._process

    def config_sample(
        self, link_obj: CaseSample, panel_bed: str | None
    ) -> dict[str, str | int | None]:
        """Return config sample data."""
        sample_data: dict[str, str | int] = self.get_sample_data(link_obj=link_obj)
        if sample_data["analysis_type"] == AnalysisType.WHOLE_GENOME_SEQUENCING:
            sample_data["capture_kit"]: str = panel_bed or DEFAULT_CAPTURE_KIT
        else:
            sample_data["capture_kit"]: str | None = panel_bed or self.get_target_bed_from_lims(
                case_id=link_obj.case.internal_id
            )
        if link_obj.mother:
            sample_data[Pedigree.MOTHER.value]: str = link_obj.mother.internal_id
        if link_obj.father:
            sample_data[Pedigree.FATHER.value]: str = link_obj.father.internal_id
        return sample_data

    def get_gene_panel(self, case_id: str, dry_run: bool = False) -> list[str]:
        """Create and return the aggregated gene panel file."""
        return self._get_gene_panel(case_id=case_id, genome_build=GENOME_BUILD_37, dry_run=dry_run)

    def get_managed_variants(self) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(genome_build=GENOME_BUILD_37)

    def get_genome_build(self, case_id: str) -> str:
        """Return the reference genome build version of a MIP-DNA analysis."""
        analysis_metadata: MipAnalysis = self.get_latest_metadata(case_id)
        return analysis_metadata.genome_build

    def get_data_analysis_type(self, case_id: str) -> str:
        """
        Return the data analysis type of a MIP-DNA analysis.
        Patch for the typical behaviour of the AnalysisAPI function.
        It does not raise an error with multiple analysis types.
        """
        case: Case = self.get_validated_case(case_id)
        analysis_types: set[str] = {
            link.sample.application_version.application.analysis_type for link in case.links
        }
        if len(analysis_types) > 1:
            LOG.warning(
                f"Multiple analysis types found. Defaulting to {AnalysisType.WHOLE_GENOME_SEQUENCING}."
            )
            return AnalysisType.WHOLE_GENOME_SEQUENCING
        return analysis_types.pop() if analysis_types else None

    def get_scout_upload_case_tags(self) -> dict:
        """Return MIP DNA Scout upload case tags."""
        return MIP_CASE_TAGS



import operator
from typing import Any, Callable

from pydantic.v1 import BaseModel, Field, validator

from cg.constants import PRECISION
from cg.exc import CgError, MetricsQCError
from cg.models.qc_metrics import QCMetrics


def _get_metric_per_sample_id(sample_id: str, metric_objs: list) -> Any:
    """Get metric for a sample_id from metric object"""
    for metric in metric_objs:
        if sample_id == metric.sample_id:
            return metric


def add_metric(name: str, values: dict) -> list[Any]:
    """Add metric to list of objects"""
    found_metrics: list = []
    raw_metrics: list = values.get("metrics_")
    metrics_validator: dict[str, Any] = values.get("metric_to_get_")
    for metric in raw_metrics:
        if name == metric.name and metric.name in metrics_validator:
            found_metrics.append(
                metrics_validator[metric.name](
                    sample_id=metric.id, step=metric.step, value=metric.value
                )
            )
    return found_metrics


def add_sample_id_metrics(parsed_metric: Any, values: dict) -> list[Any]:
    """Add parsed sample_id metrics gathered from all metrics to list"""
    sample_ids: set = values.get("sample_ids")
    sample_id_metrics: list = []
    metric_per_sample_id_map: dict = {}
    for metric_name in values.get("sample_metric_to_parse"):
        metric_per_sample_id_map.update({metric_name: values.get(metric_name)})
    for sample_id in sample_ids:
        metric_per_sample_id: dict = {"sample_id": sample_id}
        for metric_name, metric_objs in metric_per_sample_id_map.items():
            sample_metric: Any = _get_metric_per_sample_id(
                sample_id=sample_id, metric_objs=metric_objs
            )
            if sample_metric.value:
                metric_per_sample_id[metric_name]: Any = sample_metric.value
                metric_per_sample_id[metric_name + "_step"]: str = sample_metric.step
        sample_id_metrics.append(parsed_metric(**metric_per_sample_id))
    return sample_id_metrics


class MetricCondition(BaseModel):
    """Metric condition model

    Attributes:
        norm: validation condition
        threshold: validation cut off
    """

    norm: str
    threshold: float | str

    @validator("norm")
    def validate_operator(cls, norm: str) -> str:
        """Validate that an operator is accepted."""
        try:
            getattr(operator, norm)
        except AttributeError as error:
            raise CgError(
                f"{norm} is not an accepted operator for QC metric conditions."
            ) from error
        return norm


class MetricsBase(BaseModel):
    """Definition for elements in deliverables metrics file."""

    header: str | None
    id: str
    input: str
    name: str
    step: str
    value: Any
    condition: MetricCondition | None


class SampleMetric(BaseModel):
    """Define base attribute for sample metrics"""

    sample_id: str
    step: str


class MeanInsertSize(SampleMetric):
    """Definition of insert size metric"""

    value: float

    @validator("value", always=True)
    def convert_mean_insert_size(cls, value) -> int:
        """Convert raw value from float to int"""
        return int(value)


class MedianTargetCoverage(SampleMetric):
    """Definition of median target coverage"""

    value: int


class ParsedMetrics(QCMetrics):
    """Defines parsed metrics"""

    sample_id: str
    mean_insert_size: int
    mean_insert_size_step: str
    median_target_coverage: int
    median_target_coverage_step: str


class MetricsDeliverables(BaseModel):
    """Specification for a metric general deliverables file"""

    metrics_: list[MetricsBase] = Field(..., alias="metrics")
    sample_ids: set | None

    @validator("sample_ids", always=True)
    def set_sample_ids(cls, _, values: dict) -> set:
        """Set sample_ids gathered from all metrics"""
        sample_ids: list = []
        raw_metrics: list = values.get("metrics_")
        for metric in raw_metrics:
            sample_ids.append(metric.id)
        return set(sample_ids)


class MetricsDeliverablesCondition(BaseModel):
    """Specification for a metric deliverables file with conditions sets."""

    metrics: list[MetricsBase]

    @validator("metrics")
    def validate_metrics(cls, metrics: list[MetricsBase]) -> list[MetricsBase]:
        """Verify that metrics met QC conditions."""
        failed_metrics = []
        for metric in metrics:
            if metric.condition is not None:
                qc_function: Callable = getattr(operator, metric.condition.norm)
                if not qc_function(metric.value, metric.condition.threshold):
                    metric_value = (
                        round(metric.value, PRECISION)
                        if isinstance(metric.value, float)
                        else metric.value
                    )
                    failed_metrics.append(
                        f"{metric.id} - {metric.name}={metric_value} (threshold: {metric.condition.norm} {metric.condition.threshold})"
                    )
        if failed_metrics:
            raise MetricsQCError(f"QC failed: {'; '.join(failed_metrics)}")
        return metrics


class MultiqcDataJson(BaseModel):
    """Multiqc data json model."""

    report_general_stats_data: list[dict] | None
    report_data_sources: dict | None
    report_saved_raw_data: dict[str, dict] | None
